﻿using System;
using System.Text;
using System.Data;
using System.Data.SqlClient;
using System.Diagnostics;

namespace InsertPerformance
{
    class Program
    {
        delegate void InsertRecordsDelegate(SqlConnection conn, int nRecords);

        static void Main(string[] args)
        {
            using (SqlConnection conn = new SqlConnection("Server=elmserver;Database=ec2;Trusted_Connection=True;"))
            {
                conn.Open();
                DeleteOldRecords(conn);
                InsertRecordsDelegate[] methods =
                {
                    InsertRecordsOneByOne,
                    InsertRecordsNoParams,
                    InsertRecordsInTransaction,
                    InsertRecordsUsingUnionAll,
                    InsertRecordsSqlBulkCopy
                };

                int methodToUse = GetMethodToUse(args);

                Stopwatch watch = new Stopwatch();
                {
                    watch.Start();
                    int nRecords = 1000;
                    methods[methodToUse](conn, nRecords);
                    watch.Stop();
                    Console.WriteLine("Inserted {0} records in {1} ms", nRecords, watch.ElapsedMilliseconds);
                }
            }
        }

        static int GetMethodToUse(string[] args)
        {
            if (args.Length == 0) return 0;
            return int.Parse(args[0]);
        }

        static void DeleteOldRecords(SqlConnection conn)
        {
            using (SqlCommand cmd = new SqlCommand("DELETE FROM Quotes", conn))
            {
                cmd.ExecuteNonQuery();
            }
        }

        static void InsertRecordsOneByOne(SqlConnection conn, int nRecords)
        {
            Console.WriteLine("Inserting records one by one...");
            using (SqlCommand cmd = new SqlCommand("INSERT INTO Quotes (Symbol, QuoteDate, Price) VALUES (@symbol, @date, @price)", conn))
            {
                for (int i = 0; i < nRecords; ++i)
                {
                    cmd.Parameters.Clear();
                    cmd.Parameters.AddWithValue("@symbol", "MSFT");
                    cmd.Parameters.AddWithValue("@date", new DateTime(2000, 1, 1).AddDays(i));
                    cmd.Parameters.AddWithValue("@price", 12.3m);
                    cmd.ExecuteNonQuery();
                }
            }
        }

        static void InsertRecordsSqlBulkCopy(SqlConnection conn, int nRecords)
        {
            Console.WriteLine("Inserting records via SqlBulkCopy...");
            using (SqlBulkCopy copy = new SqlBulkCopy(conn))
            {
                copy.DestinationTableName = "Quotes";
                DataTable table = new DataTable("Quotes");
                table.Columns.Add("Symbol", typeof(string));
                table.Columns.Add("QuoteDate", typeof(DateTime));
                table.Columns.Add("Price", typeof(decimal));

                for (int i = 0; i < nRecords; ++i)
                {
                    table.Rows.Add("MSFT", new DateTime(2000, 1, 1).AddDays(i), 12.3m);
                }

                copy.WriteToServer(table);
            }
        }

        static void InsertRecordsInTransaction(SqlConnection conn, int nRecords)
        {
            Console.WriteLine("Inserting records in transaction...");
            using (SqlTransaction trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted))
            {
                using (SqlCommand cmd = new SqlCommand("INSERT INTO Quotes (Symbol, QuoteDate, Price) VALUES (@symbol, @date, @price)", conn, trans))
                {
                    for (int i = 0; i < nRecords; ++i)
                    {
                        cmd.Parameters.Clear();
                        cmd.Parameters.AddWithValue("@symbol", "MSFT");
                        cmd.Parameters.AddWithValue("@date", new DateTime(2000, 1, 1).AddDays(i));
                        cmd.Parameters.AddWithValue("@price", 12.3m);
                        cmd.ExecuteNonQuery();
                    }
                }

                trans.Commit();
            }
        }

        static void InsertRecordsNoParams(SqlConnection conn, int nRecords)
        {
            Console.WriteLine("Inserting records one by one (no parameters)...");
            for (int i=0; i<nRecords; ++i)
            {
                DateTime date = new DateTime(2000,1,1).AddDays(i);
                string insert = String.Format("INSERT INTO Quotes (Symbol, QuoteDate, Price) VALUES ('MSFT', {0}, 12.3)", date.ToShortDateString());

                using (SqlCommand cmd = new SqlCommand(insert, conn))
                {
                    cmd.ExecuteNonQuery();
                }
            }
        }

        static void InsertRecordsUsingUnionAll(SqlConnection conn, int nRecords)
        {
            Console.WriteLine("Inserting records using UNION ALL...");
            const int BatchSize = 500;
            for (int i = 0; i < nRecords; )
            {
                StringBuilder s = new StringBuilder();
                s.Append("INSERT INTO Quotes (Symbol, QuoteDate, Price)");

                for (int j = 0; j < BatchSize && i < nRecords; ++j, ++i)
                {
                    if (j != 0) s.Append("UNION ALL ");
                    DateTime date = new DateTime(2000, 1, 1).AddDays(i);
                    s.AppendFormat("SELECT 'MSFT', {0}, 12.3", date.ToShortDateString());
                    
                }

                using (SqlCommand cmd = new SqlCommand(s.ToString(), conn))
                {
                    cmd.ExecuteNonQuery();
                }
            }
        }

    }
}
